@@ -5,6 +5,11 @@ require 'puppet/util/puppetdb'
 class Puppet::Node::Puppetdb < Puppet::Indirector::REST
   include Puppet::Util::Puppetdb
 
+  # Run initial checks
+  def initialize
+    Puppet::Util::Puppetdb::GlobalCheck.run
+  end
+
   def find(request)
   end
 
@@ -12,6 +17,6 @@ class Puppet::Node::Puppetdb < Puppet::Indirector::REST
   end
 
   def destroy(request)
-    submit_command(request, request.key, 'deactivate node', 1)
+    submit_command(request.key, request.key, CommandDeactivateNode, 2)
   end
 end
