@@ -9,8 +9,8 @@
             [clojure.tools.logging :as log]
             [puppetlabs.kitchensink.core :as kitchensink]
             [clojure.string :as str]
-            [com.puppetlabs.time :as pl-time])
-  (:use com.puppetlabs.jdbc.internal))
+            [com.puppetlabs.time :as pl-time]
+            [com.puppetlabs.jdbc.internal :refer :all]))
 
 
 (defn valid-jdbc-query?
@@ -62,26 +62,6 @@
                      :else %)]
        (map #(kitchensink/mapvals convert %) result-set))))
 
-(defn add-limit-clause
-  "Helper function for ensuring that a query does not return more than a certain
-  number of results.  (Adds a limit clause to an SQL query if necessary.)
-
-  Accepts two parameters: `limit` and `query`.  `query` should be an SQL query
-  (String) that you wish to apply a LIMIT clause to.
-
-  `limit` is an integer specifying the maximum number of results that we are looking
-  for.  If `limit` is zero, then we return the original `query` unaltered.  If
-  `limit is greater than zero, we add a limit clause using the time-honored trick
-  of using the value of `limit + 1`;  This allows us to later compare the size of
-  the result set against the original limit and detect cases where we've exceeded
-  the maximum."
-  [limit query]
-  {:pre [(and (integer? limit) (>= limit 0))
-         (string? query)]}
-  (if (pos? limit)
-    (format "select results.* from (%s) results LIMIT %s" query (inc limit))
-    query))
-
 (defn limited-query-to-vec
   "Take a limit and an SQL query (with optional parameters), and return the
   result of the query as a vector.  These results, unlike a normal query result,
@@ -166,7 +146,9 @@
 
   Note that if no paging options are specified, the original SQL will be
   returned completely unmodified."
-  [sql {:keys [limit offset order-by]}]
+  ([sql {:keys [limit offset order-by]}]
+   (paged-sql sql {:limit limit :offset offset :order-by order-by} nil))
+  ([sql {:keys [limit offset order-by]} entity]
   {:pre [(string? sql)
          ((some-fn nil? integer?) limit)
          ((some-fn nil? integer?) offset)
@@ -175,20 +157,32 @@
    :post [(string? %)]}
     (let [limit-clause     (if limit (format " LIMIT %s" limit) "")
           offset-clause    (if offset (format " OFFSET %s" offset) "")
-          order-by-clause  (order-by->sql order-by)]
-      (format "SELECT paged_results.* FROM (%s) paged_results%s%s%s"
-          sql
-          order-by-clause
-          limit-clause
-          offset-clause)))
+          order-by-clause  (order-by->sql order-by)
+          inner-order-by   (str/replace order-by-clause #"environment"
+                                        "COALESCE(distinct_names.environment,'')")]
+      (case entity
+        :factsets
+        (format "SELECT paged_results.* FROM (%s) paged_results
+                WHERE (certname,COALESCE(paged_results.environment,''),timestamp) IN
+                (SELECT DISTINCT certname,COALESCE(distinct_names.environment,''),timestamp FROM (%s)
+                distinct_names %s%s%s) %s"
+                sql sql inner-order-by limit-clause offset-clause order-by-clause)
+        (format "SELECT paged_results.* FROM (%s) paged_results%s%s%s"
+                sql order-by-clause limit-clause offset-clause)))))
 
 (defn count-sql
   "Takes a sql string and returns a modified sql string that will select
   the count of results that would be returned by the original sql."
-  [sql]
+  ([sql]
+   (count-sql nil sql))
+  ([entity sql]
   {:pre   [(string? sql)]
    :post  [(string? %)]}
-  (format "SELECT COUNT(*) AS result_count FROM (%s) results_to_count" sql))
+  (case entity
+    :factsets
+    (format "SELECT COUNT(*) AS result_count FROM (SELECT DISTINCT certname
+            from (%s) paged_sql) results_to_count" sql)
+    (format "SELECT COUNT(*) AS result_count FROM (%s) results_to_count" sql))))
 
 (defn get-result-count
   "Takes a sql string, executes a `COUNT` statement against the database,
@@ -287,9 +281,10 @@
            stats log-statements log-slow-statements statements-cache-size
            conn-max-age conn-lifetime conn-keep-alive read-only?]
     :as   db}]
-  ;; Load the database driver class
-  (Class/forName classname)
-  (let [log-slow-statements-duration (pl-time/to-secs log-slow-statements)
+  (let [;; Load the database driver class explicitly, to avoid jar load ordering
+        ;; issues.
+        _ (Class/forName classname)
+        log-slow-statements-duration (pl-time/to-secs log-slow-statements)
         config          (doto (new BoneCPConfig)
                           (.setDefaultAutoCommit false)
                           (.setLazyInit true)
@@ -330,3 +325,14 @@
   (str "in ("
        (str/join "," (repeat (count coll) "?"))
        ")"))
+
+(defn in-clause-multi
+  "Create a prepared statement in clause, with a `width`-sized series of ? for
+  every item in coll."
+  [coll width]
+  {:pre [(seq coll)
+         (integer? width)]}
+  (let [inner (str "(" (str/join "," (repeat width "?")) ")")]
+    (str "in ("
+         (str/join "," (repeat (count coll) inner))
+         ")")))
