@@ -57,16 +57,17 @@
             [com.puppetlabs.puppetdb.config :as conf]
             [com.puppetlabs.puppetdb.utils :as utils]
             [puppetlabs.kitchensink.core :as kitchensink]
-            [puppetlabs.trapperkeeper.core :refer [defservice main]])
-  (:use [clojure.java.io :only [file]]
-        [clj-time.core :only [ago secs minutes days]]
-        [overtone.at-at :only (mk-pool interspaced)]
-        [com.puppetlabs.time :only [to-secs to-millis parse-period format-period period?]]
-        [com.puppetlabs.jdbc :only (with-transacted-connection)]
-        [com.puppetlabs.repl :only (start-repl)]
-        [com.puppetlabs.puppetdb.scf.migrate :only [migrate!]]
-        [com.puppetlabs.puppetdb.version :only [version update-info]]
-        [com.puppetlabs.puppetdb.command.constants :only [command-names]]))
+            [puppetlabs.trapperkeeper.core :refer [defservice main]]
+            [compojure.core :as compojure]
+            [clojure.java.io :refer [file]]
+            [clj-time.core :refer [ago]]
+            [overtone.at-at :refer [mk-pool interspaced]]
+            [com.puppetlabs.time :refer [to-secs to-millis parse-period format-period period?]]
+            [com.puppetlabs.jdbc :refer [with-transacted-connection]]
+            [com.puppetlabs.repl :refer [start-repl]]
+            [com.puppetlabs.puppetdb.scf.migrate :refer [migrate! indexes!]]
+            [com.puppetlabs.puppetdb.version :refer [version update-info]]
+            [com.puppetlabs.puppetdb.command.constants :refer [command-names]]))
 
 (def cli-description "Main PuppetDB daemon")
 
@@ -242,6 +243,10 @@
 
   (shutdown-mq-broker context))
 
+(defn add-max-framesize
+  "Add a maxFrameSize to broker url for activemq."
+  [config url]
+  (format "%s&wireFormat.maxFrameSize=%s&marshal=true" url (:max-frame-size config)))
 
 (defn start-puppetdb
   [context config service-id add-ring-handler shutdown-on-error]
@@ -255,8 +260,7 @@
          :as config}                            (conf/process-config! config)
         product-name                               (:product-name global)
         update-server                              (:update-server global)
-        ;; TODO: revisit the choice of 20000 as a default value for event queries
-        event-query-limit                          (:event-query-limit global)
+        url-prefix                                 (:url-prefix global)
         write-db                                   (pl-jdbc/pooled-datasource database)
         read-db                                    (pl-jdbc/pooled-datasource (assoc read-database :read-only? true))
         gc-interval                                (get database :gc-interval)
@@ -268,11 +272,11 @@
         discard-dir                                (file mq-dir "discarded")
         globals                                    {:scf-read-db          read-db
                                                     :scf-write-db         write-db
-                                                    :command-mq           {:connection-string mq-addr
+                                                    :command-mq           {:connection-string (add-max-framesize command-processing mq-addr)
                                                                            :endpoint          mq-endpoint}
-                                                    :event-query-limit    event-query-limit
                                                     :update-server        update-server
-                                                    :product-name         product-name}]
+                                                    :product-name         product-name
+                                                    :url-prefix           url-prefix}]
 
     (when (version)
       (log/info (format "PuppetDB version %s" (version))))
@@ -284,7 +288,8 @@
     ;; connection without creating anything.
     (sql/with-connection write-db
                          (scf-store/validate-database-version #(System/exit 1))
-                         (migrate!))
+                         (migrate!)
+                         (indexes! (:product-name globals)))
 
     ;; Initialize database-dependent metrics
     (pop/initialize-metrics write-db)
@@ -319,7 +324,7 @@
                                       (constantly true))
                         app (server/build-app :globals globals :authorized? authorized?)]
                     (log/info "Starting query server")
-                    (add-ring-handler app ""))
+                    (add-ring-handler (compojure/context url-prefix [] app) url-prefix))
           job-pool (mk-pool)]
 
       ;; Pretty much this helper just knows our job-pool and gc-interval
